/*
 * This file is part of Chihaya.
 *
 * Chihaya is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chihaya is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chihaya.  If not, see <http://www.gnu.org/licenses/>.
 */

package server

import (
	"bytes"
	"context"
	"fmt"
	"github.com/valyala/fasthttp"
	"math"
	"net"
	"strings"
	"time"

	"chihaya/config"
	"chihaya/database"
	cdb "chihaya/database/types"
	"chihaya/log"
	"chihaya/record"
	"chihaya/server/params"
	"chihaya/util"

	"github.com/zeebo/bencode"
)

var (
	announceInterval       int
	minAnnounceInterval    int
	peerInactivityInterval int
	maxAccounceDrift       int
	defaultNumWant         int
	maxNumWant             int

	strictPort bool

	privateIPBlocks []*net.IPNet
)

func init() {
	intervalsConfig := config.Section("intervals")
	announceConfig := config.Section("announce")

	announceInterval, _ = intervalsConfig.GetInt("announce", 1800)
	minAnnounceInterval, _ = intervalsConfig.GetInt("min_announce", 900)
	peerInactivityInterval, _ = intervalsConfig.GetInt("peer_inactivity", 3900)
	maxAccounceDrift, _ = intervalsConfig.GetInt("announce_drift", 300)

	strictPort, _ = announceConfig.GetBool("strict_port", false)
	defaultNumWant, _ = announceConfig.GetInt("numwant", 25)
	maxNumWant, _ = announceConfig.GetInt("max_numwant", 50)

	for _, cidr := range []string{
		"10.0.0.0/8",     // RFC1918
		"172.16.0.0/12",  // RFC1918
		"192.168.0.0/16", // RFC1918
		"169.254.0.0/16", // RFC3927 link-local
		"100.64.0.0/10",  // RFC6598
	} {
		_, block, err := net.ParseCIDR(cidr)
		if err != nil {
			log.Error.Printf("IP parse error on %q: %v", cidr, err)
			log.WriteStack()
		} else {
			privateIPBlocks = append(privateIPBlocks, block)
		}
	}
}

func getPublicIPV4(ipAddr string, exists bool) (string, bool) {
	if !exists { // Already does not exist, fail
		return ipAddr, exists
	}

	ip := net.ParseIP(ipAddr)
	if ip == nil { // Invalid IP provided, fail
		return ipAddr, false
	}

	if ip.To4() == nil { // IPv6 provided, fail
		return ipAddr, false
	}

	private := false

	if ip.IsLinkLocalUnicast() || ip.IsLinkLocalMulticast() {
		private = true
	} else {
		for _, block := range privateIPBlocks {
			if block.Contains(ip) {
				private = true
				break
			}
		}
	}

	return ipAddr, !private
}

func announce(
	ctx context.Context, queryArgs *fasthttp.Args, header *fasthttp.RequestHeader,
	remoteAddr net.Addr, user *cdb.User, db *database.Database, buf *bytes.Buffer) int {
	qp, err := params.ParseQuery(queryArgs)
	if err != nil {
		panic(err)
	}

	// Mandatory parameters
	infoHashes := qp.Params.InfoHashes
	peerID := qp.Params.PeerID
	port, portExists := qp.Params.Port, qp.Exists.Port
	uploaded, uploadedExists := qp.Params.Uploaded, qp.Exists.Uploaded
	downloaded, downloadedExists := qp.Params.Downloaded, qp.Exists.Downloaded
	left, leftExists := qp.Params.Left, qp.Exists.Left

	if len(infoHashes) == 0 {
		failure("Malformed request - missing info_hash", buf, 1*time.Hour)
		return fasthttp.StatusOK // Required by torrent clients to interpret failure response
	} else if len(infoHashes) > 1 {
		failure("Malformed request - can only announce singular info_hash", buf, 1*time.Hour)
		return fasthttp.StatusOK // Required by torrent clients to interpret failure response
	}

	if len(peerID) == 0 {
		failure("Malformed request - missing peer_id", buf, 1*time.Hour)
		return fasthttp.StatusOK // Required by torrent clients to interpret failure response
	}

	if len(peerID) != 20 {
		failure("Malformed request - invalid peer_id", buf, 1*time.Hour)
		return fasthttp.StatusOK // Required by torrent clients to interpret failure response
	}

	if !portExists {
		failure("Malformed request - missing port", buf, 1*time.Hour)
		return fasthttp.StatusOK // Required by torrent clients to interpret failure response
	}

	if strictPort && port < 1024 || port > 65535 {
		failure(fmt.Sprintf("Malformed request - port outside of acceptable range (port: %d)", port), buf, 1*time.Hour)
		return fasthttp.StatusOK // Required by torrent clients to interpret failure response
	}

	if !uploadedExists {
		failure("Malformed request - missing uploaded", buf, 1*time.Hour)
		return fasthttp.StatusOK // Required by torrent clients to interpret failure response
	}

	if !downloadedExists {
		failure("Malformed request - missing downloaded", buf, 1*time.Hour)
		return fasthttp.StatusOK // Required by torrent clients to interpret failure response
	}

	if !leftExists {
		failure("Malformed request - missing left", buf, 1*time.Hour)
		return fasthttp.StatusOK // Required by torrent clients to interpret failure response
	}

	ipAddr := func() string {
		ipV4, existsV4 := getPublicIPV4(qp.Params.IPv4, qp.Exists.IPv4) // First try to get IPv4 address if client sent it
		ip, exists := getPublicIPV4(qp.Params.IP, qp.Exists.IP)         // ... then try to get public IP if sent by client

		// Fail if ip and ipv4 are not same, and both are provided
		if (existsV4 && exists) && (ip != ipV4) {
			return ""
		}

		if existsV4 {
			return ipV4
		}

		if exists {
			return ip
		}

		// Check for proxied IP header
		if proxyHeader, exists := config.Section("http").Get("proxy_header", ""); exists {
			if ips := header.PeekAll(proxyHeader); len(ips) > 0 {
				return string(ips[0])
			}
		}

		// Check for IP in socket
		remoteAddrString := remoteAddr.String()
		portIndex := strings.LastIndex(remoteAddrString, ":")

		if portIndex != -1 {
			return remoteAddrString[0:portIndex]
		}

		// Everything else failed
		return ""
	}()

	ipBytes := net.ParseIP(ipAddr).To4()
	if nil == ipBytes {
		failure(fmt.Sprintf("Failed to parse IP address (ip: %s)", ipAddr), buf, 1*time.Hour)
		return fasthttp.StatusOK // Required by torrent clients to interpret failure response
	}

	clientID, matched := clientApproved(peerID, db)
	if !matched {
		failure(fmt.Sprintf("Your client is not approved (peer_id: %s)", peerID), buf, 1*time.Hour)
		return fasthttp.StatusOK // Required by torrent clients to interpret failure response
	}

	torrent, exists := (*db.Torrents.Load())[infoHashes[0]]
	if !exists {
		failure("This torrent does not exist", buf, 5*time.Minute)
		return fasthttp.StatusOK // Required by torrent clients to interpret failure response
	}

	if torrentStatus := torrent.Status.Load(); torrentStatus == 1 && left == 0 {
		log.Info.Printf("Unpruning torrent %d", torrent.ID.Load())

		torrent.Status.Store(0)

		/* It is okay to do this asynchronously as tracker's internal in-memory state has already been updated for this
		torrent. While it is technically possible that we will do this more than once in some cases, the state is of
		boolean type so there is no risk of data loss. */
		go db.UnPrune(torrent)
	} else if torrentStatus != 0 {
		failure(fmt.Sprintf("This torrent does not exist (status: %d, left: %d)", torrentStatus, left), buf, 15*time.Minute)
		return fasthttp.StatusOK // Required by torrent clients to interpret failure response
	}

	numWant, exists := qp.Params.NumWant, qp.Exists.NumWant
	if !exists {
		numWant = uint16(defaultNumWant)
	} else if numWant > uint16(maxNumWant) {
		numWant = uint16(maxNumWant)
	}

	var (
		peer    *cdb.Peer
		now     = time.Now().Unix()
		newPeer = false
		seeding = false
		active  = true
	)

	event := qp.Params.Event
	completed := event == "completed"

	userID := user.ID.Load()
	peerKey := cdb.NewPeerKey(userID, cdb.PeerIDFromRawString(peerID))

	// Take torrent peers lock to read/write on it to prevent race conditions
	if !torrent.TryLockWithContext(ctx) {
		return fasthttp.StatusRequestTimeout
	}
	defer torrent.Unlock()

	if left > 0 {
		if isDisabledDownload(db, user, torrent) {
			failure("Your download privileges are disabled", buf, 1*time.Hour)
			return fasthttp.StatusOK // Required by torrent clients to interpret failure response
		}

		peer, exists = torrent.Leechers[peerKey]
		if !exists {
			newPeer = true
			peer = &cdb.Peer{}
			torrent.Leechers[peerKey] = peer

			torrent.LeechersLength.Store(uint32(len(torrent.Leechers)))
		}
	} else if completed {
		peer, exists = torrent.Leechers[peerKey]
		if !exists {
			newPeer = true
			peer = &cdb.Peer{}
			torrent.Seeders[peerKey] = peer

			torrent.SeedersLength.Store(uint32(len(torrent.Seeders)))
		} else {
			// Previously tracked peer is now a seeder
			torrent.Seeders[peerKey] = peer
			delete(torrent.Leechers, peerKey)

			torrent.SeedersLength.Store(uint32(len(torrent.Seeders)))
			torrent.LeechersLength.Store(uint32(len(torrent.Leechers)))
		}
		seeding = true
	} else {
		peer, exists = torrent.Seeders[peerKey]
		if !exists {
			peer, exists = torrent.Leechers[peerKey]
			if !exists {
				newPeer = true
				peer = &cdb.Peer{}
				torrent.Seeders[peerKey] = peer

				torrent.SeedersLength.Store(uint32(len(torrent.Seeders)))
			} else {
				/* Previously tracked peer is now a seeder, however we never received their "completed" event.
				Broken client? Unreported snatch? Cross-seeding? Let's not report it as snatch to avoid
				over-reporting for cross-seeding */
				torrent.Seeders[peerKey] = peer
				delete(torrent.Leechers, peerKey)

				torrent.SeedersLength.Store(uint32(len(torrent.Seeders)))
				torrent.LeechersLength.Store(uint32(len(torrent.Leechers)))
			}
		}
		seeding = true
	}

	// Update peer info/stats
	if newPeer {
		peer.StartTime = now
		peer.RelativeLastAnnounce = 0
		peer.Uploaded = uploaded
		peer.Downloaded = downloaded
	}

	rawDeltaUpload := int64(uploaded) - int64(peer.Uploaded)
	rawDeltaDownload := int64(downloaded) - int64(peer.Downloaded)

	// If a user restarts a torrent, their delta may be negative, attenuating this to 0 should be fine for stats purposes
	if rawDeltaUpload < 0 {
		rawDeltaUpload = 0
	}

	if rawDeltaDownload < 0 {
		rawDeltaDownload = 0
	}

	torrentGroupDownMultiplier := 1.0
	torrentGroupUpMultiplier := 1.0

	if torrentGroupFreeleech, exists := (*db.TorrentGroupFreeleech.Load())[torrent.Group.Key()]; exists {
		torrentGroupDownMultiplier = torrentGroupFreeleech.DownMultiplier
		torrentGroupUpMultiplier = torrentGroupFreeleech.UpMultiplier
	}

	var deltaDownload int64
	if !database.GlobalFreeleech.Load() {
		deltaDownload = int64(float64(rawDeltaDownload) * math.Abs(math.Float64frombits(user.DownMultiplier.Load())) *
			math.Abs(torrentGroupDownMultiplier) * math.Abs(math.Float64frombits(torrent.DownMultiplier.Load())))
	}

	deltaUpload := int64(float64(rawDeltaUpload) * math.Abs(math.Float64frombits(user.UpMultiplier.Load())) *
		math.Abs(torrentGroupUpMultiplier) * math.Abs(math.Float64frombits(torrent.UpMultiplier.Load())))
	peer.Uploaded = uploaded
	peer.Downloaded = downloaded
	peer.Left = left
	peer.Seeding = seeding
	deltaTime := now - peer.LastAnnounce()

	if deltaTime > int64(peerInactivityInterval) {
		deltaTime = 0
	}

	var deltaSeedTime int64
	if seeding {
		deltaSeedTime = now - peer.LastAnnounce()
	}

	if deltaSeedTime > int64(peerInactivityInterval) {
		deltaSeedTime = 0
	}

	peer.SetLastAnnounce(now)
	/* Update torrent last_action only if announced action is seeding.
	This allows dead torrents without seeder but with leecher to be proeprly pruned */
	if seeding {
		torrent.LastAction.Store(now)
	}

	var deltaSnatch uint8

	if event == "stopped" {
		/* We can remove the peer from the list and still have their stats be recorded,
		since we still have a reference to their object. After flushing, all references
		should be gone, allowing the peer to be GC'd. */
		if seeding {
			delete(torrent.Seeders, peerKey)
			torrent.SeedersLength.Store(uint32(len(torrent.Seeders)))
		} else {
			delete(torrent.Leechers, peerKey)
			torrent.LeechersLength.Store(uint32(len(torrent.Leechers)))
		}

		active = false
	} else if completed {
		db.QueueSnatch(userID, torrent.ID.Load(), now)
		deltaSnatch = 1
	}

	if user.TrackerHide.Load() {
		peer.Addr = cdb.NewPeerAddressFromIPPort(net.IP{127, 0, 0, 1}, port) // 127.0.0.1
	} else {
		peer.Addr = cdb.NewPeerAddressFromIPPort(ipBytes, port)
	}

	peer.ClientID = clientID

	// If the channels are already full, block until a flush occurs
	db.QueueTorrent(torrent, deltaSnatch)
	db.QueueTransferHistory(
		peer, userID, torrent.ID.Load(), rawDeltaUpload,
		rawDeltaDownload, deltaTime, deltaSeedTime, deltaSnatch, active)
	db.QueueUser(user, rawDeltaUpload, rawDeltaDownload, deltaUpload, deltaDownload)
	db.QueueTransferIP(peer, userID, torrent.ID.Load(), rawDeltaUpload, rawDeltaDownload)

	go record.Record(
		torrent.ID.Load(),
		userID,
		ipAddr,
		port,
		event,
		seeding,
		rawDeltaUpload,
		rawDeltaDownload,
		uploaded,
		downloaded,
		left)

	// Generate response
	seedCount := int(torrent.SeedersLength.Load())
	leechCount := int(torrent.LeechersLength.Load())
	snatchCount := uint16(torrent.Snatched.Load())

	response := make(map[string]interface{})
	response["complete"] = seedCount
	response["incomplete"] = leechCount
	response["downloaded"] = snatchCount
	response["min interval"] = minAnnounceInterval

	/* We ask clients to announce each interval seconds. In order to spread the load on tracker,
	we will vary the interval given to client by random number of seconds between 0 and value
	specified in config */
	announceDrift := util.UnsafeFastRand(0, maxAccounceDrift)
	response["interval"] = announceInterval + announceDrift

	if numWant > 0 && active {
		compact := !qp.Exists.Compact || !qp.Params.Compact

		noPeerID := qp.Exists.NoPeerID && qp.Params.NoPeerID

		var peerCount int
		if seeding {
			peerCount = util.Min(int(numWant), leechCount)
		} else {
			peerCount = util.Min(int(numWant), leechCount+seedCount-1)
		}

		type peerToSend struct {
			Peer *cdb.Peer
			Key  cdb.PeerKey
		}

		peersToSend := make([]peerToSend, 0, peerCount)

		/*
		 * The iteration is already "random", so we don't need to randomize ourselves:
		 * - Each time an element is inserted into the map, it gets a some arbitrary position for iteration
		 * - Each time you range over the map, it starts at a random offset into the map's elements
		 */
		if seeding {
			for k, leech := range torrent.Leechers {
				if len(peersToSend) >= int(numWant) {
					break
				}

				if k.ID() == userID {
					continue
				}

				peersToSend = append(peersToSend, peerToSend{
					Peer: leech,
					Key:  k,
				})
			}
		} else {
			/* Send only one peer per user. This is to ensure that users seeding at multiple locations don't end up
			exclusively acting as peers. */
			uniqueSeeders := make(map[uint32]struct{})
			for k, seed := range torrent.Seeders {
				if len(peersToSend) >= int(numWant) {
					break
				}

				seedUserID := k.ID()

				if seedUserID == userID {
					continue
				}

				if _, exists = uniqueSeeders[seedUserID]; !exists {
					uniqueSeeders[seedUserID] = struct{}{}
					peersToSend = append(peersToSend, peerToSend{
						Peer: seed,
						Key:  k,
					})
				}
			}
			for k, leech := range torrent.Leechers {
				if len(peersToSend) >= int(numWant) {
					break
				}
				if k.ID() == userID {
					continue
				}
				peersToSend = append(peersToSend, peerToSend{
					Peer: leech,
					Key:  k,
				})
			}
		}

		if compact {
			peerBuff := make([]byte, 0, len(peersToSend)*cdb.PeerAddressSize)

			for _, other := range peersToSend {
				peerBuff = append(peerBuff, other.Peer.Addr[:]...)
			}

			response["peers"] = peerBuff
		} else {
			peerList := make([]map[string]interface{}, len(peersToSend))
			for i, other := range peersToSend {
				peerMap := make(map[string]interface{})
				peerMap["ip"] = other.Peer.Addr.IPString()
				peerMap["port"] = other.Peer.Addr.Port()
				if !noPeerID {
					thisPeerID := other.Key.PeerID()
					peerMap["peer id"] = thisPeerID[:]
				}
				peerList[i] = peerMap
			}
			response["peers"] = peerList
		}
	}

	// Early exit before response write
	select {
	case <-ctx.Done():
		return fasthttp.StatusRequestTimeout
	default:
	}

	encoder := bencode.NewEncoder(buf)
	if err := encoder.Encode(response); err != nil {
		panic(err)
	}

	return fasthttp.StatusOK
}
