/*
 * This file is part of Chihaya.
 *
 * Chihaya is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chihaya is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chihaya.  If not, see <http://www.gnu.org/licenses/>.
 */

// Package params is based on https://github.com/chihaya/chihaya/blob/e6e7269/bittorrent/params.go
package params

import (
	"bytes"
	cdb "chihaya/database/types"
	"github.com/valyala/fasthttp"
	"strconv"
)

type QueryParam struct {
	Params struct {
		Uploaded   uint64
		Downloaded uint64
		Left       uint64

		Port    uint16
		NumWant uint16

		PeerID string
		IPv4   string
		IP     string
		Event  string

		TestGarbageUnescape string

		Compact  bool
		NoPeerID bool

		InfoHashes []cdb.TorrentHash
	}

	Exists struct {
		Uploaded   bool
		Downloaded bool
		Left       bool

		Port    bool
		NumWant bool

		PeerID bool
		IPv4   bool
		IP     bool
		Event  bool

		TestGarbageUnescape bool

		Compact  bool
		NoPeerID bool

		InfoHashes bool
	}
}

var uploadedKey = []byte("uploaded")
var downloadedKey = []byte("downloaded")
var leftKey = []byte("left")

var portKey = []byte("port")
var numWant = []byte("numwant")

var peerIDKey = []byte("peer_id")
var ipv4Key = []byte("ipv4")
var ipKey = []byte("ip")
var eventKey = []byte("event")

var testGarbageUnescapeKey = []byte("!@#")

var infoHashKey = []byte("info_hash")

var compactKey = []byte("compact")
var noPeerIDKey = []byte("no_peer_id")

func ParseQuery(queryArgs *fasthttp.Args) (qp QueryParam, err error) {
	var returnError error

	queryArgs.VisitAll(func(key, value []byte) {
		if returnError != nil {
			return
		}

		key = bytes.ToLower(key)
		switch true {
		case bytes.Equal(key, uploadedKey):
			n, err := strconv.ParseUint(string(value), 10, 0)
			if err != nil {
				returnError = err
				return
			}
			qp.Params.Uploaded = n
			qp.Exists.Uploaded = true
		case bytes.Equal(key, downloadedKey):
			n, err := strconv.ParseUint(string(value), 10, 0)
			if err != nil {
				returnError = err
				return
			}
			qp.Params.Downloaded = n
			qp.Exists.Downloaded = true
		case bytes.Equal(key, leftKey):
			n, err := strconv.ParseUint(string(value), 10, 0)
			if err != nil {
				returnError = err
				return
			}
			qp.Params.Left = n
			qp.Exists.Left = true
		case bytes.Equal(key, portKey):
			n, err := strconv.ParseUint(string(value), 10, 16)
			if err != nil {
				returnError = err
				return
			}
			qp.Params.Port = uint16(n)
			qp.Exists.Port = true
		case bytes.Equal(key, numWant):
			n, err := strconv.ParseUint(string(value), 10, 16)
			if err != nil {
				returnError = err
				return
			}
			qp.Params.NumWant = uint16(n)
			qp.Exists.NumWant = true
		case bytes.Equal(key, peerIDKey):
			qp.Params.PeerID = string(value)
			qp.Exists.PeerID = true
		case bytes.Equal(key, ipv4Key):
			qp.Params.IPv4 = string(value)
			qp.Exists.IPv4 = true
		case bytes.Equal(key, ipKey):
			qp.Params.IP = string(value)
			qp.Exists.IP = true
		case bytes.Equal(key, eventKey):
			qp.Params.Event = string(value)
			qp.Exists.Event = true
		case bytes.Equal(key, testGarbageUnescapeKey):
			qp.Params.TestGarbageUnescape = string(value)
			qp.Exists.TestGarbageUnescape = true
		case bytes.Equal(key, infoHashKey):
			if len(value) == cdb.TorrentHashSize {
				qp.Params.InfoHashes = append(qp.Params.InfoHashes, cdb.TorrentHashFromBytes(value))
				qp.Exists.InfoHashes = true
			}
		case bytes.Equal(key, compactKey):
			qp.Params.Compact = bytes.Equal(value, []byte{'1'})
			qp.Exists.Compact = true
		case bytes.Equal(key, noPeerIDKey):
			qp.Params.NoPeerID = bytes.Equal(value, []byte{'1'})
			qp.Exists.NoPeerID = true
		}
	})

	return qp, returnError
}
