/*
 * This file is part of Chihaya.
 *
 * Chihaya is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chihaya is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chihaya.  If not, see <http://www.gnu.org/licenses/>.
 */

package server

import (
	"bytes"
	"context"
	"github.com/valyala/fasthttp"
	"net"
	"path"
	"sync"
	"sync/atomic"
	"time"

	"chihaya/collectors"
	"chihaya/config"
	"chihaya/database"
	"chihaya/log"
	"chihaya/record"
	"chihaya/util"

	"github.com/prometheus/client_golang/prometheus"
)

type httpHandler struct {
	startTime time.Time

	normalRegisterer prometheus.Registerer
	normalCollector  *collectors.NormalCollector
	adminCollector   *collectors.AdminCollector

	bufferPool *util.BufferPool

	db *database.Database

	requests   atomic.Uint64
	throughput int

	contextTimeout time.Duration

	waitGroup sync.WaitGroup
	terminate bool
}

var (
	handler  *httpHandler
	listener net.Listener
)

func (handler *httpHandler) respond(ctx context.Context, requestContext *fasthttp.RequestCtx, buf *bytes.Buffer) int {
	dir, action := path.Split(string(requestContext.Request.URI().Path()))
	if action == "" {
		return fasthttp.StatusNotFound
	}

	/*
	 * ===================================================
	 * Handle public endpoints (/:action)
	 * ===================================================
	 */

	passkey := path.Dir(dir)[1:]
	if passkey == "" {
		switch action {
		case "alive":
			return alive(buf)
		}

		return fasthttp.StatusNotFound
	}

	/*
	 * ===================================================
	 * Handle private endpoints (/:passkey/:action)
	 * ===================================================
	 */

	user := isPasskeyValid(passkey, handler.db)
	if user == nil {
		failure("Your passkey is invalid", buf, 1*time.Hour)
		return fasthttp.StatusOK
	}

	switch action {
	case "announce":
		return announce(
			ctx,
			requestContext.Request.URI().QueryArgs(),
			&requestContext.Request.Header,
			requestContext.RemoteAddr(),
			user,
			handler.db,
			buf)
	case "scrape":
		if enabled, _ := config.GetBool("scrape", true); !enabled {
			return fasthttp.StatusNotFound
		}

		return scrape(ctx, requestContext.Request.URI().QueryArgs(), user, handler.db, buf)
	case "metrics":
		return metrics(ctx, requestContext.Request.Header.PeekBytes([]byte("Authorization")), handler.db, buf)
	}

	return fasthttp.StatusNotFound
}

func (handler *httpHandler) handler(requestContext *fasthttp.RequestCtx) {
	if handler.terminate {
		return
	}

	// Count new request (done before everything else so that failed/timeout numbers match)
	handler.requests.Add(1)

	// Mark request as being handled, so that server won't shutdown before we're done with it
	handler.waitGroup.Add(1)
	defer handler.waitGroup.Done()

	buf := handler.bufferPool.Take()
	// Mark buf to be returned to bufferPool after we are done with it
	defer handler.bufferPool.Give(buf)

	// Gracefully handle panics so that they're confined to single request and don't crash server
	defer func() {
		if err := recover(); err != nil {
			log.Error.Printf("Recovered from panicking request handler - %v\nURL was: %s", err, requestContext.URI().String())
			log.WriteStack()

			collectors.IncrementErroredRequests()

			if len(requestContext.Response.Header.ContentType()) == 0 {
				buf.Reset()
				requestContext.Response.SetStatusCode(fasthttp.StatusInternalServerError)
			}
		}
	}()

	// Prepare and start new context with timeout to abort long-running requests
	ctx, cancel := context.WithTimeout(context.Background(), handler.contextTimeout)
	defer cancel()

	/* Pass flow to handler; note that handler should be responsible for actually cancelling
	its own work based on request context cancellation */
	status := handler.respond(ctx, requestContext, buf)

	select {
	case <-ctx.Done():
		switch err := ctx.Err(); err {
		case context.DeadlineExceeded:
			collectors.IncrementTimeoutRequests()

			failure("Request context deadline exceeded", buf, 5*time.Minute)

			requestContext.Response.Header.SetContentLength(buf.Len())
			requestContext.Response.Header.SetContentTypeBytes([]byte("text/plain"))
			requestContext.Response.SetStatusCode(fasthttp.StatusOK) // Required by torrent clients to interpret failure response
			_, _ = requestContext.Write(buf.Bytes())
		case context.Canceled:
			collectors.IncrementCancelRequests()

			requestContext.Response.SetStatusCode(fasthttp.StatusRequestTimeout)
		}
	default:
		requestContext.Response.Header.SetContentLength(buf.Len())
		requestContext.Response.Header.SetContentTypeBytes([]byte("text/plain"))
		requestContext.Response.SetStatusCode(status)
		_, _ = requestContext.Write(buf.Bytes())
	}
}

func Start() {
	handler = &httpHandler{db: &database.Database{}, startTime: time.Now()}

	/* Initialize reusable buffer pool; this is faster than allocating new memory for every request.
	If necessary, new memory will be allocated when pool is empty, however. */
	bufferPool := util.NewBufferPool(512)
	handler.bufferPool = bufferPool

	addr, _ := config.Section("http").Get("addr", ":34000")
	readTimeout, _ := config.Section("http").Section("timeout").GetInt("read", 1)
	writeTimeout, _ := config.Section("http").Section("timeout").GetInt("write", 3)
	idleTimeout, _ := config.Section("http").Section("timeout").GetInt("idle", 30)

	// Set appropriate context timeout based on write timeout; 200ms should be enough as a wiggle room
	handler.contextTimeout = time.Duration(writeTimeout)*time.Second - 200*time.Millisecond

	// Create new server instance
	server := &fasthttp.Server{
		Handler:                      handler.handler,
		ReadTimeout:                  time.Duration(readTimeout) * time.Second,
		WriteTimeout:                 time.Duration(writeTimeout) * time.Second,
		IdleTimeout:                  time.Duration(idleTimeout) * time.Second,
		GetOnly:                      true,
		DisablePreParseMultipartForm: true,
		NoDefaultServerHeader:        true,
		NoDefaultDate:                true,
		NoDefaultContentType:         true,
		CloseOnShutdown:              true,
	}

	if idleTimeout <= 0 {
		log.Warning.Print("Setting idleTimeout <= 0 disables Keep-Alive which might negatively impact performance")

		server.DisableKeepalive = true
	}

	// Start new goroutine to calculate throughput
	go func() {
		lastTime := time.Now()
		lastRequests := handler.requests.Load()

		for !handler.terminate {
			time.Sleep(time.Minute)

			now := time.Now()
			duration := now.Sub(lastTime)
			requests := handler.requests.Load()

			handler.throughput = int(float64(requests-lastRequests)/duration.Seconds()*60 + 0.5)
			log.Info.Printf("Current throughput: %d rpm", handler.throughput)

			lastTime = now
			lastRequests = requests
		}
	}()

	// Initialize database and recorder
	handler.db.Init()
	record.Init()

	// Register default prometheus collector
	handler.normalRegisterer = prometheus.NewRegistry()
	handler.normalCollector = collectors.NewNormalCollector()
	handler.normalRegisterer.MustRegister(handler.normalCollector)

	// Register additional metrics for DefaultGatherer
	handler.adminCollector = collectors.NewAdminCollector()
	prometheus.MustRegister(handler.adminCollector)

	// Start TCP listener
	var err error

	listener, err = net.Listen("tcp", addr)
	if err != nil {
		panic(err)
	}

	log.Info.Printf("Ready and accepting new connections on %s", addr)

	/* Start serving new request. Behind the scenes, this works by spawning a new goroutine for each client.
	This is pretty fast and scalable since goroutines are nice and efficient. Blocks until TCP listener is closed. */
	_ = server.Serve(listener)

	// Wait for active connections to finish processing
	handler.waitGroup.Wait()

	_ = server.Shutdown()

	log.Info.Print("Now closed and not accepting any new connections")

	// Close database connection
	handler.db.Terminate()

	log.Info.Print("Shutdown complete")
}

func Stop() {
	if listener != nil {
		// Closing the listener stops accepting connections and causes Serve to return
		_ = listener.Close()
	}

	handler.terminate = true
}
