/*
 * This file is part of Chihaya.
 *
 * Chihaya is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chihaya is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chihaya.  If not, see <http://www.gnu.org/licenses/>.
 */

package util

import (
	"math/rand"
	"testing"
)

func TestUnsafeRand(t *testing.T) {
	for i := 0; i < 10; i++ {
		min := rand.Intn(1000)
		max := rand.Intn(1000) + min
		randomInt := UnsafeRand(min, max)

		if randomInt < min || randomInt > max {
			t.Fatalf("Integer %d is outside specified range (%d - %d)", randomInt, min, max)
		}
	}
}

func TestUnsafeFastDistribution(t *testing.T) {
	var slots [901]uint64

	for i := 0; i < len(slots)*100000; i++ {
		slots[UnsafeFastRand(0, len(slots)-1)]++
	}

	var total uint64
	for k := range slots {
		total += slots[k]
	}

	for k := range slots {
		expected := 100 / float64(len(slots))
		actual := (float64(slots[k]) / float64(total)) * 100
		diff := actual - expected
		t.Logf("%3d: actual %f%%, expected %f%%, diff %f%%", k, actual, expected, diff)

		if diff > 0.01 {
			t.Error("out of distribution")
		}
	}
}

func TestUnsafeFastRand(t *testing.T) {
	for i := 0; i < 10; i++ {
		min := rand.Intn(1000)
		max := rand.Intn(1000) + min
		randomInt := UnsafeFastRand(min, max)

		if randomInt < min || randomInt > max {
			t.Fatalf("Integer %d is outside specified range (%d - %d)", randomInt, min, max)
		}
	}
}

func TestUnsafeIntn(t *testing.T) {
	for i := 1; i < 2000; i++ {
		genInt := UnsafeIntn(i)

		if genInt < 0 || genInt >= i {
			t.Fatalf("Generated random integer (%d) does not fall in the range [0, %d)!", genInt, i)
		}
	}
}
